@@ -40,12 +40,8 @@
 #define DOORBELL_SIZE(gfxv)	(((gfxv) >= 0x90000) ? 8 : 4)
 #define DOORBELLS_PAGE_SIZE(ds)	(1024 * (ds))
 
-#define EOP_BUFFER_SIZE(gfxv)					\
-	(((gfxv) == GFX_VERSION_TONGA) ? TONGA_PAGE_SIZE :	\
-	(((gfxv) >= 0x80000) ? 4096 : 0))
-
 #define WG_CONTEXT_DATA_SIZE_PER_CU(gfxv) 		\
-	(VGPR_SIZE_PER_CU(gfxv) + SGPR_SIZE_PER_CU +	\
+	(get_vgpr_size_per_cu(gfxv) + SGPR_SIZE_PER_CU +	\
 	 LDS_SIZE_PER_CU + HWREG_SIZE_PER_CU)
 
 #define CNTL_STACK_BYTES_PER_WAVE(gfxv)	\
@@ -66,6 +62,7 @@ struct queue {
 	uint32_t ctl_stack_size;
 	uint32_t debug_memory_size;
 	uint32_t eop_buffer_size;
+	uint32_t total_mem_alloc_size;
 	uint32_t gfxv;
 	bool use_ats;
 	bool unified_ctx_save_restore;
@@ -87,6 +84,22 @@ struct process_doorbells {
 static unsigned int num_doorbells;
 static struct process_doorbells *doorbells;
 
+uint32_t get_vgpr_size_per_cu(uint32_t gfxv)
+{
+	uint32_t vgpr_size = 0x40000;
+
+	if ((gfxv & ~(0xff)) == GFX_VERSION_AQUA_VANJARAM ||
+		 gfxv == GFX_VERSION_ALDEBARAN ||
+		 gfxv == GFX_VERSION_ARCTURUS)
+		vgpr_size = 0x80000;
+
+	else if (gfxv == GFX_VERSION_PLUM_BONITO ||
+		 gfxv == GFX_VERSION_WHEAT_NAS)
+		vgpr_size = 0x60000;
+
+	return vgpr_size;
+}
+
 HSAKMT_STATUS init_process_doorbells(unsigned int NumNodes)
 {
 	unsigned int i;
@@ -270,7 +283,7 @@ static bool update_ctx_save_restore_size(uint32_t nodeid, struct queue *q)
 		return false;
 	if (node.NumFComputeCores && node.NumSIMDPerCU) {
 		uint32_t ctl_stack_size, wg_data_size;
-		uint32_t cu_num = node.NumFComputeCores / node.NumSIMDPerCU;
+		uint32_t cu_num = node.NumFComputeCores / node.NumSIMDPerCU / node.NumXcc;
 		uint32_t wave_num = (q->gfxv < GFX_VERSION_NAVI10)
 			? MIN(cu_num * 40, node.NumShaderBanks / node.NumArrays * 512)
 			: cu_num * 32;
@@ -279,8 +292,7 @@ static bool update_ctx_save_restore_size(uint32_t nodeid, struct queue *q)
 		wg_data_size = cu_num * WG_CONTEXT_DATA_SIZE_PER_CU(q->gfxv);
 		q->ctl_stack_size = PAGE_ALIGN_UP(sizeof(HsaUserContextSaveAreaHeader)
 					+ ctl_stack_size);
-		if (q->gfxv >= GFX_VERSION_NAVI10 &&
-		    q->gfxv <= GFX_VERSION_YELLOW_CARP) {
+		if ((q->gfxv & 0x3f0000) == 0xA0000) {
 			/* HW design limits control stack size to 0x7000.
 			 * This is insufficient for theoretical PM4 cases
 			 * but sufficient for AQL, limited by SPI events.
@@ -401,7 +413,7 @@ static HSAKMT_STATUS register_svm_range(void *mem, uint32_t size,
 		flags |= HSA_SVM_FLAG_GPU_ALWAYS_MAPPED;
 	}
 
-	nattr = 5;
+	nattr = 6;
 	s_attr = sizeof(*attrs) * nattr;
 	attrs = (HSA_SVM_ATTRIBUTE *)alloca(s_attr);
 
@@ -415,6 +427,8 @@ static HSAKMT_STATUS register_svm_range(void *mem, uint32_t size,
 	attrs[3].value = flags;
 	attrs[4].type = HSA_SVM_ATTR_ACCESS;
 	attrs[4].value = gpuNode;
+	attrs[5].type = HSA_SVM_ATTR_GRANULARITY;
+	attrs[5].value = 0xFF;
 
 	return hsaKmtSVMSetAttr(mem, size, nattr, attrs);
 }
@@ -426,27 +440,31 @@ static void free_queue(struct queue *q)
 					 q->eop_buffer_size,
 					 PAGE_SIZE, q->use_ats);
 	if (q->unified_ctx_save_restore)
-		free(q->ctx_save_restore);
+		munmap(q->ctx_save_restore, q->total_mem_alloc_size);
 	else if (q->ctx_save_restore)
 		free_exec_aligned_memory(q->ctx_save_restore,
-					 q->ctx_save_restore_size,
+					 q->total_mem_alloc_size,
 					 PAGE_SIZE, q->use_ats);
 
 	free_exec_aligned_memory((void *)q, sizeof(*q), PAGE_SIZE, q->use_ats);
 }
 
 static inline void fill_cwsr_header(struct queue *q, void *addr,
-		HsaEvent *Event, volatile HSAint64 *ErrPayload)
+		HsaEvent *Event, volatile HSAint64 *ErrPayload, HSAuint32 NumXcc)
 {
-	HsaUserContextSaveAreaHeader *header =
-			(HsaUserContextSaveAreaHeader *)addr;
+	uint32_t i;
+	HsaUserContextSaveAreaHeader *header;
 
-	header->ErrorEventId = 0;
-	if (Event)
-		header->ErrorEventId = Event->EventId;
-	header->ErrorReason = ErrPayload;
-	header->DebugOffset = q->ctx_save_restore_size;
-	header->DebugSize = q->debug_memory_size;
+	for (i = 0; i < NumXcc; i++) {
+		header = (HsaUserContextSaveAreaHeader *)
+			((uintptr_t)addr + (i * q->ctx_save_restore_size));
+		header->ErrorEventId = 0;
+		if (Event)
+			header->ErrorEventId = Event->EventId;
+		header->ErrorReason = ErrPayload;
+		header->DebugOffset = (NumXcc - i) * q->ctx_save_restore_size;
+		header->DebugSize = q->debug_memory_size * NumXcc;
+	}
 }
 
 static int handle_concrete_asic(struct queue *q,
@@ -475,36 +493,49 @@ static int handle_concrete_asic(struct queue *q,
 	ret = update_ctx_save_restore_size(NodeId, q);
 
 	if (ret) {
-		uint32_t total_mem_alloc_size = 0;
 		HsaNodeProperties node;
 		bool svm_api;
 
+		if (hsaKmtGetNodeProperties(NodeId, &node))
+			return HSAKMT_STATUS_ERROR;
+
 		args->ctx_save_restore_size = q->ctx_save_restore_size;
 		args->ctl_stack_size = q->ctl_stack_size;
 
 		/* Total memory to be allocated is =
-		 * (Control Stack size + WG size) + Debug memory area size
+		 * (Control Stack size + WG size +
+		 *  Debug memory area size) * num_xcc
 		 */
-		total_mem_alloc_size = q->ctx_save_restore_size +
-				       q->debug_memory_size;
+		q->total_mem_alloc_size = (q->ctx_save_restore_size +
+					q->debug_memory_size) * node.NumXcc;
 
-		if (hsaKmtGetNodeProperties(NodeId, &node))
-			svm_api = false;
-		else
-			svm_api = node.Capability.ui32.SVMAPISupported;
+		svm_api = node.Capability.ui32.SVMAPISupported &&
+			  (node.EngineId.ui32.Major < 11 || node.DeviceId == 0x7461);
 
 		/* Allocate unified memory for context save restore
 		 * area on dGPU.
 		 */
 		if (!q->use_ats && svm_api) {
-			uint32_t size = PAGE_ALIGN_UP(total_mem_alloc_size);
+			uint32_t size = PAGE_ALIGN_UP(q->total_mem_alloc_size);
 			void *addr;
 			HSAKMT_STATUS r = HSAKMT_STATUS_ERROR;
 
-			if (posix_memalign(&addr, GPU_HUGE_PAGE_SIZE, size))
-				pr_err("[%s] posix_memalign failed:\n", __func__);
-			else {
-				fill_cwsr_header(q, addr, Event, ErrPayload);
+			addr = mmap_allocate_aligned(PROT_READ | PROT_WRITE,
+						     MAP_ANONYMOUS | MAP_PRIVATE,
+						     size, GPU_HUGE_PAGE_SIZE, 0,
+						     0, (void *)LONG_MAX);
+			if (!addr) {
+				pr_err("mmap failed to alloc ctx area size 0x%x: %s\n",
+					size, strerror(errno));
+			} else {
+				/*
+				 * To avoid fork child process COW MMU notifier
+				 * callback evict parent process queues.
+				 */
+				if (madvise(addr, size, MADV_DONTFORK))
+					pr_err("madvise failed -%d\n", errno);
+
+				fill_cwsr_header(q, addr, Event, ErrPayload, node.NumXcc);
 
 				r = register_svm_range(addr, size,
 						NodeId, NodeId, 0, true);
@@ -512,21 +543,22 @@ static int handle_concrete_asic(struct queue *q,
 				if (r == HSAKMT_STATUS_SUCCESS) {
 					q->ctx_save_restore = addr;
 					q->unified_ctx_save_restore = true;
-				} else
-					free(addr);
+				} else {
+					munmap(addr, size);
+				}
 			}
 		}
 
 		if (!q->unified_ctx_save_restore) {
 			q->ctx_save_restore = allocate_exec_aligned_memory(
-							total_mem_alloc_size,
+							q->total_mem_alloc_size,
 							q->use_ats, NodeId,
 							false, false, false);
 
 			if (!q->ctx_save_restore)
 				return HSAKMT_STATUS_NO_MEMORY;
 
-			fill_cwsr_header(q, q->ctx_save_restore, Event, ErrPayload);
+			fill_cwsr_header(q, q->ctx_save_restore, Event, ErrPayload, node.NumXcc);
 		}
 
 		args->ctx_save_restore_address = (uintptr_t)q->ctx_save_restore;
@@ -577,7 +609,13 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtCreateQueue(HSAuint32 NodeId,
 
 	q->gfxv = get_gfxv_by_node_id(NodeId);
 	q->use_ats = false;
-	q->eop_buffer_size = EOP_BUFFER_SIZE(q->gfxv);
+
+	if (q->gfxv == GFX_VERSION_TONGA)
+		q->eop_buffer_size = TONGA_PAGE_SIZE;
+	else if ((q->gfxv & ~(0xff)) == GFX_VERSION_AQUA_VANJARAM)
+		q->eop_buffer_size = ((Type == HSA_QUEUE_COMPUTE) ? 4096 : 0);
+	else if (q->gfxv >= 0x80000)
+		q->eop_buffer_size = 4096;
 
 	/* By default, CUs are all turned on. Initialize cu_mask to '1
 	 * for all CU bits.
@@ -641,16 +679,16 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtCreateQueue(HSAuint32 NodeId,
 	q->queue_id = args.queue_id;
 
 	if (IS_SOC15(q->gfxv)) {
+		HSAuint64 mask = DOORBELLS_PAGE_SIZE(DOORBELL_SIZE(q->gfxv)) - 1;
+
 		/* On SOC15 chips, the doorbell offset within the
 		 * doorbell page is included in the doorbell offset
 		 * returned by KFD. This allows CP queue doorbells to be
 		 * allocated dynamically (while SDMA queue doorbells fixed)
 		 * rather than based on the its process queue ID.
 		 */
-		doorbell_mmap_offset = args.doorbell_offset &
-			~(HSAuint64)(doorbells[NodeId].size - 1);
-		doorbell_offset = args.doorbell_offset &
-			(doorbells[NodeId].size - 1);
+		doorbell_mmap_offset = args.doorbell_offset & ~mask;
+		doorbell_offset = args.doorbell_offset & mask;
 	} else {
 		/* On older chips, the doorbell offset within the
 		 * doorbell page is based on the queue ID.
@@ -662,7 +700,6 @@ HSAKMT_STATUS HSAKMTAPI hsaKmtCreateQueue(HSAuint32 NodeId,
 	err = map_doorbell(NodeId, gpu_id, doorbell_mmap_offset);
 	if (err != HSAKMT_STATUS_SUCCESS) {
 		hsaKmtDestroyQueue(q->queue_id);
-		free_queue(q);
 		return HSAKMT_STATUS_ERROR;
 	}
 
